@@ -1,4 +1,4 @@
-from typing import Iterator, List
+from typing import Iterator
 
 from .name import Name
 from .rdata import Rdata
@@ -8,7 +8,6 @@ class RRset:
     name: Name
     rdtype: int
 
-    def __iter__(self) -> Iterator[Rdata]:
-        ...
+    def __iter__(self) -> Iterator[Rdata]: ...
 
     ...
