@@ -2,17 +2,30 @@ import asyncio
 import functools
 import logging
 import os
+import sys
+from collections.abc import Callable, Coroutine
+
+if sys.version_info >= (3, 10):
+    from typing import ParamSpec
+else:
+    from typing_extensions import ParamSpec
+
+from aioice import ice
+
+P = ParamSpec("P")
 
 
-def asynctest(coro):
+def asynctest(
+    coro: Callable[P, Coroutine[None, None, None]],
+) -> Callable[P, None]:
     @functools.wraps(coro)
-    def wrap(*args, **kwargs):
+    def wrap(*args: P.args, **kwargs: P.kwargs) -> None:
         asyncio.run(coro(*args, **kwargs))
 
     return wrap
 
 
-async def invite_accept(conn_a, conn_b):
+async def invite_accept(conn_a: ice.Connection, conn_b: ice.Connection) -> None:
     # invite
     await conn_a.gather_candidates()
     for candidate in conn_a.local_candidates:
@@ -30,7 +43,7 @@ async def invite_accept(conn_a, conn_b):
     conn_a.remote_password = conn_b.local_password
 
 
-def read_message(name):
+def read_message(name: str) -> bytes:
     path = os.path.join(os.path.dirname(__file__), "data", name)
     with open(path, "rb") as fp:
         return fp.read()
